/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.mahout.clustering.fuzzykmeans;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.WritableComparable;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.mahout.clustering.ClusterObservations;
import org.apache.mahout.math.VectorWritable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class FuzzyKMeansMapper extends Mapper<WritableComparable<?>,VectorWritable,Text,ClusterObservations> {
  
  private static final Logger log = LoggerFactory.getLogger(FuzzyKMeansMapper.class);
  
  private final List<SoftCluster> clusters = new ArrayList<SoftCluster>();
  private FuzzyKMeansClusterer clusterer;

  @Override
  protected void map(WritableComparable<?> key, VectorWritable point, Context context)
    throws IOException, InterruptedException {
    clusterer.emitPointProbToCluster(point.get(), clusters, context);
  }

  @Override
  protected void setup(Context context) throws IOException, InterruptedException {
    super.setup(context);
    Configuration job = context.getConfiguration();
    clusterer = new FuzzyKMeansClusterer(job);
    
    log.info("In Mapper Configure:");
    
    String clusterPath = job.get(FuzzyKMeansConfigKeys.CLUSTER_PATH_KEY);
    if (clusterPath != null && clusterPath.length() > 0) {
      FuzzyKMeansUtil.configureWithClusterInfo(new Path(clusterPath), clusters);
    }
    
    if (clusters.isEmpty()) {
      throw new IllegalStateException("No clusters found. Check your -c path.");
    }
  }

  /**
   * Configure the mapper by providing its clusters. Used by unit tests.
   * 
   * @param clusters
   *          a List<Cluster>
   */
  void config(Collection<SoftCluster> clusters) {
    this.clusters.clear();
    this.clusters.addAll(clusters);
  }  
}
